/*
  Macro           SortOnRegExp
  Author          Carlo Hogeveen
  Website         ecarlo.nl/tse
  Compatibility   TSE v4 upwards, all TSE variants
  Version         v1   24 May 2023

  This tool sorts the lines of a buffer or block, either on a regular
  expression's matches, or on a sort key of regular expression tags.

  Syntax:

    SortOnRegExp [-i] <regular expression> [sort key]

  If no block exists in the current buffer, then the whole buffer will be
  sorted.

  If a block exists in the current buffer, then both these rules apply:
  - The tool will only sort lines in which the block occurs.
  - The tool will only find matches which completely occur inside the block.

  You can provide parameters through the macro command line,
  otherwise the tool will provide a parameter prompt.
  The parameters must be provided in the documented order.

  The options parameter is optional.

    Default a sort will be case-sensitive.
    With option parameter "-i" a sort will be case-insensitive.

  The regular expression parameter is mandatory.

    It must be a TSE regular expression.
    The regular expression must to be quoted if it contains any spaces.

    See "Searching With Regular Expressions" in TSE's Help for how TSE's
    regular expressions work.

  The sort key parameter is optional.

    If you do not provide a sort key, then each line's matched string becomes
    its sort key.
    If a selected line has no match, then a "bottom value" becomes its sort
    key, sorting non-matching lines to the bottom.

    The sort key parameter is a string containing tag references, which refer
    to tagged strings in the regular expression.

    A tag reference has the format "\<digit>", where <digit> is a character
    from "1" through "9".
    So a sort key could look like \1\3\2 or "\1 \3 \2" or \1~\3~\2.
    In the current version of the tool it makes sense to separate tag
    references by a character with a high ASCII code, like "~" or "".

    See "{ }" in "Searching With Regular Expressions" in TSE's Help
    for how TSE's regular expressions can use tags and tag-references.

  Lines with a match will be sorted above lines without a match.
  This is wrong and will be reversed in a next version of the tool.

  Lines with the same match retain their relative positions.


  TODO

  Because I think that in practice this tool will almost never be used,
  I decided to not develop it further until someone finds a good reason.


  THEREFORE NOT TODO
    MUST
    - The tool is too slow to sort a larger number of lines.
    - The tool should not remove the undo history.
    - Sort unmatched lines above matched lines.
    SHOULD
    - Remove the need for a separator character between tag references.
    - Syntax to sort a tag numerically, e.g. sort "5" before "10".
    COULD
    - A reverse-sort option "r".
    - Without the regular expression do a whole line or block sort.
    WONT


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there, for example by
  opening it there and applying the Macro Compile menu.

  As a tool, "SortOnRegExp" is a good candidate to add to TSE's Potpourri menu.


  HISTORY

  v0.0.0.1   17 Apr 2023
    It works as a proof of concept, but see the TODO list.

  v0.0.0.2   22 Apr 2023
    Main change: The documentation was improved a LOT!
    Made the tool TSE v4 upwards compatiple for all TSE variants.

  v1         24 May 2023
    Only changed the version number to indicate,
    that lacking a good reason I will not develop this tool any further.
*/





// Start of compatibility restrictions and mitigations


/*
  When compiled with a TSE version below TSE 4.0 the compiler reports
  a syntax error and hilights the first applicable line below.

*/

/*
  About 16-bits vs 32-bits and Windows vs Linux detection.

  There is a beta Linux version of TSE that is not 100% bug-free and in which
  some features do not work, but all its Linux versions are above TSE 4.0,
  and they all are 32-bits which is what WIN32 actually signifies,
  so for Linux we should define the constant WIN32 but as FALSE.

  So then, as many macros already do, we can still use
    #ifdef WIN32
  to determine if the environment is 32-bits,
  but then we can also use the clearer
    #if WIN32
  as an alternative to
    #ifndef LINUX
  to determine if the environment is Windows.
*/

#ifdef LINUX
  #define WIN32 FALSE
#endif

#ifdef WIN32
#else
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifdef EDITOR_VERSION
#else
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif



#if EDITOR_VERSION < 4400h
  /*
    StrFind() 1.0

    If you have TSE Pro 4.0 or 4.2, then this proc implements the core of the
    built-in StrFind() function of TSE Pro 4.4.
    The StrFind() function searches a string or pattern inside another string
    and returns the position of the found string or zero.
    It works for strings like the regular Find() function does for files,
    so read the Help for the regular Find() function for the usage of the
    options, but apply these differences while reading:
    - Where the Find() (related) documentation refers to "file" and "line",
      StrFind() refers to "string".
    - The search option "g" ("global", meaning "from the start of the string")
      is implicit and can therefore always be omitted.
    As with the regular Find() function all characters are allowed as options,
    but here only these are acted upon: b, i, w, x, ^, $.

    Notable differences between the procedure below with TSE 4.4's built-in
    function:
    - The third parameter "options" is mandatory.
    - No fourth parameter "start" (actually occurrence: which one to search).
    - No fifth  parameter "len" (returning the length of the found text).

    Technical implementation notes:
    - To be reuseable elsewhere the procedure's source code is written to work
      independently of the rest of the source code.
      That said, it is intentionally not implemented as an include file, both
      for ease of installation and because one day another macro might need its
      omitted parameters, which would be an include file nightmare.
    - A tiny downside of the independent part is, that StrFind's buffer is not
      purged with the macro. To partially compensate for that if the macro is
      restarted, StrFind's possibly pre-existing buffer is searched for.
    - The fourth and fifth parameter are not implemented.
      - The first reason was that I estimated the tiny but actual performance
        gain and the easier function call to be more beneficial than the
        slight chance of a future use of these extra parameters.
      - The main reason turned out to be that in TSE 4.4 the fourth parameter
        "start" is erroneously documented and implemented.
        While this might be corrected in newer versions of TSE, it neither
        makes sense to me to faithfully reproduce these errors here, nor to
        make a correct implementation that will be replaced by an incorrect
        one if you upgrade to TSE 4.4.
  */
  integer strfind_id = 0
  integer proc StrFind(string needle, string haystack, string options)
    integer i                           = 0
    string  option                  [1] = ''
    integer org_id                      = GetBufferId()
    integer result                      = FALSE  // Zero.
    string  strfind_name [MAXSTRINGLEN] = ''
    string  validated_options       [7] = 'g'
    for i = 1 to Length(options)
      option = Lower(SubStr(options, i, 1))
      if      (option in 'b', 'i', 'w', 'x', '^', '$')
      and not Pos(option, validated_options)
        validated_options = validated_options + option
      endif
    endfor
    if strfind_id
      GotoBufferId(strfind_id)
      EmptyBuffer()
    else
      strfind_name = SplitPath(CurrMacroFilename(), _NAME_) + ':StrFind'
      strfind_id   = GetBufferId(strfind_name)
      if strfind_id
        GotoBufferId(strfind_id)
        EmptyBuffer()
      else
        strfind_id = CreateTempBuffer()
        ChangeCurrFilename(strfind_name, _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      endif
    endif
    InsertText(haystack, _INSERT_)
    if lFind(needle, validated_options)
      result = CurrPos()
    endif
    GotoBufferId(org_id)
    return(result)
  end StrFind
#endif


#if EDITOR_VERSION < 4500h
  integer proc SwapLines(integer p_line1, integer p_line2)
    integer line1                     = p_line1
    integer line2                     = p_line2
    integer old_InsertLineBlocksAbove = 0
    integer result                    = FALSE

    if line1 > line2
      line1 = p_line2
      line2 = p_line1
    endif

    if  line1 >  0
    and line2 >  line1
    and line2 <= NumLines()
      PushBlock()
      PushLocation()
      old_InsertLineBlocksAbove = Set(InsertLineBlocksAbove, OFF)

      GotoLine(line2)
      MarkLine(CurrLine(), CurrLine())
      GotoLine(line1)
      MoveBlock()
      MarkLine(CurrLine(), CurrLine())
      GotoLine(line2)
      MoveBlock()

      Set(InsertLineBlocksAbove, old_InsertLineBlocksAbove)
      PopLocation()
      PopBlock()
    endif

    return(result)
  end SwapLines
#endif


// End of compatibility restrictions and mitigations.





// Constants and semi-constants
string MACRO_NAME  [MAXSTRINGLEN] = ''
string LAST_STRING [MAXSTRINGLEN] = ''


/*
  The following function determines the minimum amount of characters that a
  given regular expression can match. For the practical purpose that they are
  valid matches too, "^" and "$" are pretended to have length 1.
  The purpose of this procedure is, to be able to beforehand avoid searching
  for an empty string, which is logically pointless because it always succeeds
  (with one exception: past the end of the last line).
  Searching for an empty string in a loop makes it infinate and hangs a macro.
  Properly applied, using this procedure can avoid that.
*/
integer proc minimum_regexp_length(string s)
  integer addition           = 0
  integer i                  = 0
  integer NEXT_TIME          = 2
  integer orred_addition     = 0
  integer prev_addition      = 0
  integer prev_i             = 0
  integer result             = 0
  integer tag_level          = 0
  integer THIS_TIME          = 1
  integer use_orred_addition = FALSE
  // For each character.
  for i = 1 to Length(s)
    // Ignore this zero-length "character".
    if Lower(SubStr(s,i,2)) == "\c"
      i = i + 1
    else
      // Skip index for all these cases so they will be counted as one
      // character.
      case SubStr(s,i,1)
        when "["
          while i < Length(s)
          and   SubStr(s,i,1) <> "]"
            i = i + 1
          endwhile
        when "\"
          i = i + 1
          case Lower(SubStr(s,i,1))
            when "x"
              i = i + 2
            when "d", "o"
              i = i + 3
          endcase
      endcase
      // Now start counting.
      if use_orred_addition == NEXT_TIME
         use_orred_addition =  THIS_TIME
      endif
      // Count a literal character as one:
      if SubStr(s,i-1,1) == "\" // (Using the robustness of SubStr!)
        addition = 1
      // Count a tagged string as the length of its subexpression:
      elseif SubStr(s,i,1) == "{"
        prev_i = i
        tag_level = 1
        while i < Length(s)
        and   (tag_level <> 0 or SubStr(s,i,1) <> "}")
          i = i + 1
          if SubStr(s,i,1) == "{"
            tag_level = tag_level + 1
          elseif SubStr(s,i,1) == "}"
            tag_level = tag_level - 1
          endif
        endwhile
        addition = minimum_regexp_length(SubStr(s,prev_i+1,i-prev_i-1))
      // For a "previous character or tagged string may occur zero
      // times" operator: since it was already counted, subtract it.
      elseif SubStr(s,i,1) in "*", "@", "?"
        addition = -1 * Abs(prev_addition)
      // This is a tough one: the "or" operator.
      // For now subtract the length of the previous character or
      // tagged string, but remember doing so, because you might have
      // to add it again instead of the length of the character or
      // tagged string after the "or" operator.
      elseif SubStr(s,i,1) == "|"
        addition           = -1 * Abs(prev_addition)
        orred_addition     = Abs(prev_addition)
        use_orred_addition = NEXT_TIME
      else
      // Count ordinary characters as 1 character.
        addition = 1
      endif
      if use_orred_addition == THIS_TIME
        if orred_addition < addition
          addition = orred_addition
        endif
        use_orred_addition = FALSE
      endif
      result        = result + addition
      prev_addition = addition
    endif
  endfor
  return(result)
end minimum_regexp_length


string proc get_found_text(string tagstr)
  integer i                        = 1
  string  result    [MAXSTRINGLEN] = ''
  string  tag_match [MAXSTRINGLEN] = ''
  integer tag_nr                   = 0

  if tagstr == ''
    result = GetFoundText()
  else
    result = tagstr
    while i < Length(result)
      if result[i] == '\'
        tag_nr = Val(result[i + 1])
        if (tag_nr in 1 .. 9)
          tag_match = GetFoundText(tag_nr)
          result = result[1: i - 1] +
                         tag_match +
                         result[i + 2: MAXSTRINGLEN]
          i            = i - 2 + Length(tag_match)  // Avoid recursion.
        endif
      endif
      i = i + 1
    endwhile
  endif

  return(result)
end get_found_text


proc sort_lines_on_regexp(string regexp, string tagstr, string case_option,
                          integer line_from, integer line_to)
  integer finished               = (line_from >= line_to)
  integer loops                  = 0
  string  match_1 [MAXSTRINGLEN] = ''
  string  match_2 [MAXSTRINGLEN] = ''
  string  search_options     [5] = ''
  integer old_UndoMode           = OFF

  search_options = 'cgx' + case_option + iif(isBlockInCurrFile(), 'l', '')

  PushBlock()
  PushLocation()
  old_UndoMode = not(not(UndoMode(OFF)))

  while not finished
  and       loops <= line_to - line_from + 1
    finished = TRUE
    GotoLine(line_from)
    while CurrLine() < line_to
      match_1 = iif(lFind(regexp, search_options), get_found_text(tagstr), LAST_STRING)
      Down()
      match_2 = iif(lFind(regexp, search_options), get_found_text(tagstr), LAST_STRING)
      if match_1 > match_2
        SwapLines(CurrLine() - 1, CurrLine())
        finished = FALSE
      endif
    endwhile
    loops = loops + 1
  endwhile

  PopLocation()
  PopBlock()
  UndoMode(old_UndoMode)
end sort_lines_on_regexp


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME  = SplitPath(CurrMacroFilename(), _NAME_)
  LAST_STRING = Format(Chr(255):255:Chr(255))
end WhenLoaded


proc Main()
  string  case_option       [1] = ''
  integer line_from             = 0
  integer line_to               = 0
  integer ok                    = TRUE
  string  regexp [MAXSTRINGLEN] = Query(MacroCmdLine)
  string  tagstr [MAXSTRINGLEN] = ''

  if isBlockInCurrFile()
    line_from = Query(BlockBegLine)
    line_to   = Query(BlockEndLine)
  else
    line_from = 1
    line_to   = NumLines()
  endif

  if regexp == ''
  and not Ask('[-i] <regular expression> [tag string to sort on]',
              regexp,
              GetFreeHistory(MACRO_NAME + ':RegExp'))
    ok = FALSE
  endif

  if ok
    if EquiStr(GetFileToken(regexp, 1), '-i')
      case_option = 'i'
      regexp      = regexp[Pos('-i', Lower(regexp)) + 1: MAXSTRINGLEN]
    endif

    case NumFileTokens(regexp)
      when 0
        Warn(MACRO_NAME; 'error: No regexp provided.')
      when 1
        ok = ok
      when 2
        if  StrFind('\{.*\}' , GetFileToken(regexp, 1), 'x')
        and StrFind('\\[0-9]', GetFileToken(regexp, 2), 'x')
          tagstr = GetFileToken(regexp, 2)
          regexp = GetFileToken(regexp, 1)
        else
          Warn(MACRO_NAME; 'error:', Chr(13),
               'The parameters are not a regular expression and a tag string.', Chr(13),
               'Quote a regular expression that contains spaces', Chr(13),
               'or use a correct regular expression and tag string.')
          ok = FALSE
        endif
      otherwise
        Warn(MACRO_NAME; 'error: Too many parameters.', Chr(13),
             'Quote any strings that contain spaces.')
        ok = FALSE
    endcase
  endif

  if ok
    if minimum_regexp_length(regexp)

      if line_from < line_to
        sort_lines_on_regexp(regexp, tagstr, case_option, line_from, line_to)
      endif
    else
      Warn(MACRO_NAME; 'error:', Chr(13),
           'The regular expression matches an empty string,', Chr(13),
           'which would pointlessly match all lines.')
    endif
  endif

  PurgeMacro(CurrMacroFilename())
end Main

